#!/usr/bin/env python
'''Python3.2 unit testing framework for FW4EX based on unittest.

This framework brings new features. (1) successful assertions are
counted. (2) numbers of successful tests, failures, assertions is
printed in a bash-friendly format. (3) partial results are flushed in
a file as soon as known (this allows the tests to be killed even if
unfinished and still get the latest results).

Simple usage:
   
     import fw4exunittest

     class MyTestCase(fw4exunittest.TestCase):
       def test_1(self):
         # almost similar code as in unittest.TestCase except that, in 
         # order to not reveal the values that are expected in assertions, 
         # you should not write:
         assertEqual(computation(1,2,3), 42

         # but rather:
         expected = 42
         assertEqual(computation(1,2,3), expected)

         # This is so because, for a failing assertion, the last line of the
         # text of the assertion is printed and knowing what value to return
         # allows student to easily cheat!

         # Remember also that a testXX method stops when an assertion is not
         # verified but other tests will be executed. 
         # Remember also that testXX methods are ran in alphabetic order.

     if __name__ == '__main__':
         fw4exunittest.main()

Start the tests with:

     python test_something.py --result /tmp/lr.sh

and watch the results in /tmp/lr.sh (default is ./lastResults.sh).
The unittest module will fill stdout and stderr as usual except that
some parts of the traceback section are removed in order to prevent
leaking the values that are expected in the tests.

Historically, this module was written for Python3.4 which does not work
with Python3.2. The main changes are related to the parsing of arguments
within TestProgram (see below).

'''

__author__ = "Christian Queinnec"
__email__ = "christian dot queinnec at paracamplus dot com"
__version__ = "0.1"

import unittest
import argparse
import warnings
import time
import traceback
import re
import os

class TextTestResult(unittest.TextTestResult):
    """Holder for test result information.
     
    Just add to the parent unittest.TestResult
       - an attribute to record the number of attempted assertions
       - a method to stringify the results in a shell friendly way.

    CAUTION: this class needs its 'resultfile' attribute. This attribute 
    is set by TextTestRunner._makeResult()

    """
    def __init__(self, stream=None, descriptions=None, verbosity=1):
        unittest.TextTestResult.__init__(self, 
             stream=stream, 
             descriptions=descriptions, 
             verbosity=verbosity)
        self.assertions = 0
        self.successful_assertions = 0

    def __repr__(self):
        s = "<%s run=%i errors=%i failures=%i assertions=%i/%i resultfile=%s>" % \
         (str(self.__class__), self.testsRun, len(self.errors),
          len(self.failures), 
          self.successful_assertions, 
          self.assertions, 
          self.resultfile)
        return s

    def generateResults(self):
        s = "#/bin/bash\n"
        s += "#number of TESTS attempted:\n"
        s += "TESTS=%d\n" % self.testsRun
        failed, errored = map(len, (self.failures, self.errors))
        s += "#number of failed TESTS:\n"
        s += "FAILURES=%d\n" % (failed + errored)
        # errored ???
        s += "#number of attempted assertions:\n" 
        s += "ATTEMPTED_ASSERTIONS=%d\n" % self.assertions
        s += "#number of successful assertions:\n"
        s += "ASSERTIONS=%d\n" % self.successful_assertions
        return s

    def writeResults(self):
        #import sys; sys.stdout.write(str(self) + "\n")   # DEBUG
        s = self.generateResults()
        with open(self.resultfile, "w") as out:
            out.write(s)

    def _exc_info_to_string(self, err, test):
        """Converts a sys.exc_info()-style tuple of values into a string."""
        exctype, value, tb = err
        # Skip test runner traceback levels
        while tb and self._is_relevant_tb_level(tb):
            tb = tb.tb_next

        if exctype is test.failureException:
            # Skip assert*() traceback levels
            length = self._count_relevant_tb_levels(tb)
            msgLines = traceback.format_exception(exctype, value, tb, length)
        else:
            msgLines = traceback.format_exception(exctype, value, tb)
        
        # remove information revealing values expected from assertions:
        keptLines = []
        for item in msgLines:
            if re.match('^Assertion.*', item):
                pass
            elif re.search('/fw4exunittest.py', item):
                break
            else:
                keptLines.append(item)
        msgLines = keptLines

        if self.buffer:
            output = sys.stdout.getvalue()
            error = sys.stderr.getvalue()
            if output:
                if not output.endswith('\n'):
                    output += '\n'
                msgLines.append(STDOUT_LINE % output)
            if error:
                if not error.endswith('\n'):
                    error += '\n'
                msgLines.append(STDERR_LINE % error)
        return ''.join(msgLines)


class TextTestRunner(unittest.TextTestRunner):
    """A test runner class that displays results in textual form.

    """
    def __init__(self, stream=None, 
                 descriptions=True, verbosity=1,
                 failfast=False, buffer=False, 
                 resultclass=TextTestResult, warnings=None):
        unittest.TextTestRunner.__init__(self, 
            stream=stream, 
            descriptions=descriptions, # singular in 3.4
            verbosity=verbosity,
            failfast=failfast, 
            buffer=buffer, 
            resultclass=resultclass, 
            warnings=warnings)
        self.resultclass = TextTestResult
        self.result = None

    def _makeResult(self):
        if self.result is None:
            result = self.resultclass(self.stream, 
                                      self.descriptions, 
                                      self.verbosity)
            result.resultfile = self.resultfile
            self.result = result
        return self.result

    def run(self, test):
        "Run the given test case or test suite."
        # (1) create the result now (rather than in super().run(test)):
        self.result = self._makeResult()

        # (2) Patch all TestCases so they know this result
        # Here, test is a TestSuite of TestCases or TestSuites
        def fix(item):
            if isinstance(item, TestCase):
                item.testresult = self.result
            elif isinstance(item, unittest.TestSuite):
                for index, subitem in enumerate(item):
                    fix(subitem)
        fix(test)
        
        # (3) run as before:
        self.result = super().run(test);
        self.result.writeResults()
        return self.result

# NOTA: TestCase needs a TextTestResult to record the number of
# assertions, while TextTestRunner needs to read that TextTestResult
# to generate the final results of all tests. It must be the same object.

class TestCase(unittest.TestCase):
    """An extended TestCase class.

    CAUTION: this class needs a 'testresult' attribute holding the
    TestResult object (holding the 'assertions' and
    'successful_assertions' attributes to increment).

    """
    def __init__(self, methodName='runTest'):
        self.testresult = None
        unittest.TestCase.__init__(self, methodName=methodName)

    def recordAssertions(self):
        if self.testresult is not None:
            # Count all assertions:
            self.testresult.assertions += 1

    def recordResults(self):
        if self.testresult is not None:
            # Count successful assertions:
            self.testresult.successful_assertions += 1
            self.testresult.writeResults()

    def FW4EXassert(assertName):
        def wrap(f):
            def wrapped(self, *args):
                self.recordAssertions()
                f(self, *args)
                self.recordResults()
            return wrapped
        return wrap(unittest.TestCase.__dict__[assertName])

    assertFalse = FW4EXassert('assertFalse')
    assertTrue  = FW4EXassert('assertTrue')
    assertRaises = FW4EXassert('assertRaises')
    #assertWarns = FW4EXassert('assertWarns')
    assertEqual = FW4EXassert('assertEqual')
    assertNotEqual = FW4EXassert('assertNotEqual')
    assertAlmostEqual = FW4EXassert('assertAlmostEqual')
    assertNotAlmostEqual = FW4EXassert('assertNotAlmostEqual')
    assertSequenceEqual = FW4EXassert('assertSequenceEqual')
    assertListEqual = FW4EXassert('assertListEqual')
    assertTupleEqual = FW4EXassert('assertTupleEqual')
    assertSetEqual = FW4EXassert('assertSetEqual')
    assertIn = FW4EXassert('assertIn')
    assertNotIn = FW4EXassert('assertNotIn')
    assertIs = FW4EXassert('assertIs')
    assertIsNot = FW4EXassert('assertIsNot')
    assertDictEqual = FW4EXassert('assertDictEqual')
    assertDictContainsSubset = FW4EXassert('assertDictContainsSubset')
    #assertCountEqual = FW4EXassert('assertCountEqual')
    assertMultiLineEqual = FW4EXassert('assertMultiLineEqual')
    assertLess = FW4EXassert('assertLess')
    assertLessEqual = FW4EXassert('assertLessEqual')
    assertGreater = FW4EXassert('assertGreater')
    assertGreaterEqual = FW4EXassert('assertGreaterEqual')
    assertIsNone = FW4EXassert('assertIsNone')
    assertIsNotNone = FW4EXassert('assertIsNotNone')
    assertIsInstance = FW4EXassert('assertIsInstance')
    assertNotIsInstance = FW4EXassert('assertNotIsInstance')
    #assertRaisesRegex = FW4EXassert('assertRaisesRegex')
    #assertWarnsRegex = FW4EXassert('assertWarnsRegex')
    #assertRegex = FW4EXassert('assertRegex')
    #assertNotRegex = FW4EXassert('assertNotRegex')

class TestProgram(unittest.TestProgram):
    """Extended command-line program.

This command-line program adds a new switch:
    -r, --resultfile     file where to store tests results
where the bash-friendly result of the tests will be stored after every
assertion check.
    """
    def __init__(self, module='__main__', defaultTest=None, argv=None,
                 testRunner=None, 
                 testLoader=unittest.loader.defaultTestLoader,
                 exit=True, verbosity=1, failfast=None, catchbreak=None,
                 buffer=None, warnings=None):
        self.resultfile = 'lastResults.sh'
        unittest.TestProgram.__init__(self, 
           module=module, 
           defaultTest=defaultTest, 
           argv=argv,
           testRunner=testRunner, 
           testLoader=testLoader,
           exit=exit, 
           verbosity=verbosity, 
           failfast=failfast, 
           catchbreak=catchbreak,
           buffer=buffer, 
           warnings=warnings)

    def runTests(self):
        #import sys; sys.stdout.write(str(self))  # DEBUG
        self.testRunner = TextTestRunner(verbosity=self.verbosity,
                                         failfast=self.failfast,
                                         buffer=self.buffer,
                                         warnings=self.warnings)
        self.testRunner.resultfile = self.resultfile
        # This will trigger self.testRunner.run(self):
        unittest.TestProgram.runTests(self)

    def parseArgs(self, argv):
        if ((len(argv) > 1 and argv[1].lower() == 'discover') or
            (len(argv) == 1 and self.module is None)):
            self._do_discovery(argv[2:])
            return

        import getopt
        long_opts = ['help', 'verbose', 'quiet', 'failfast', 'catch', 'buffer',
                     'resultfile=' ]
        try:
            options, args = getopt.getopt(argv[1:], 'hHvqfcbr:', long_opts)
        except getopt.error as msg:
            self.usageExit(msg)
            return

        for opt, value in options:
            if opt in ('-h','-H','--help'):
                self.usageExit()
            if opt in ('-q','--quiet'):
                self.verbosity = 0
            if opt in ('-v','--verbose'):
                self.verbosity = 2
            if opt in ('-f','--failfast'):
                if self.failfast is None:
                    self.failfast = True
                # Should this raise an exception if -f is not valid?
            if opt in ('-c','--catch'):
                if self.catchbreak is None:
                    self.catchbreak = True
                # Should this raise an exception if -c is not valid?
            if opt in ('-b','--buffer'):
                if self.buffer is None:
                    self.buffer = True
                # Should this raise an exception if -b is not valid?
            if opt in ('-r', '--resultfile'):
                self.resultfile = value

        if len(args) == 0 and self.module is None:
            # this allows "python -m unittest -v" to still work for
            # test discovery. This means -c / -b / -v / -f options will
            # be handled twice, which is harmless but not ideal.
            self._do_discovery(argv[1:])
            return

        if len(args) == 0 and self.defaultTest is None:
            # createTests will load tests from self.module
            self.testNames = None
        elif len(args) > 0:
            self.testNames = _convert_names(args)
            if __name__ == '__main__':
                # to support python -m unittest ...
                self.module = None
        else:
            self.testNames = (self.defaultTest,)
        self.createTests()

    def _do_discovery(self, argv, Loader=unittest.loader.TestLoader):
        # handle command line args for test discovery
        self.progName = '%s discover' % self.progName
        import optparse
        parser = optparse.OptionParser()
        parser.prog = self.progName
        parser.add_option('-v', '--verbose', dest='verbose', default=False,
                          help='Verbose output', action='store_true')
        if self.failfast != False:
            parser.add_option('-f', '--failfast', dest='failfast', default=False,
                              help='Stop on first fail or error',
                              action='store_true')
        if self.catchbreak != False:
            parser.add_option('-c', '--catch', dest='catchbreak', default=False,
                              help='Catch ctrl-C and display results so far',
                              action='store_true')
        if self.buffer != False:
            parser.add_option('-b', '--buffer', dest='buffer', default=False,
                              help='Buffer stdout and stderr during tests',
                              action='store_true')
        parser.add_option('-s', '--start-directory', dest='start', default='.',
                          help="Directory to start discovery ('.' default)")
        parser.add_option('-p', '--pattern', dest='pattern', default='test*.py',
                          help="Pattern to match tests ('test*.py' default)")
        parser.add_option('-t', '--top-level-directory', dest='top', default=None,
                          help='Top level directory of project (defaults to start directory)')
        parser.add_option('-r', '--resultfile',
                          dest='resultfile',
                          default='lastResults.sh',
                          help='Filename where to store partial results')

        options, args = parser.parse_args(argv)
        if len(args) > 3:
            self.usageExit()

        for name, value in zip(('start', 'pattern', 'top'), args):
            setattr(options, name, value)

        # only set options from the parsing here
        # if they weren't set explicitly in the constructor
        if self.failfast is None:
            self.failfast = options.failfast
        if self.catchbreak is None:
            self.catchbreak = options.catchbreak
        if self.buffer is None:
            self.buffer = options.buffer

        if options.verbose:
            self.verbosity = 2

        start_dir = options.start
        pattern = options.pattern
        top_level_dir = options.top

        loader = Loader()
        self.test = loader.discover(start_dir, pattern, top_level_dir)

    def __repr__(self):
        return ("<%s resultfile=%s>" %
               (str(self.__class__), self.resultfile))

# Verbatim copy from /usr/lib/python3.2/unittest/main.py
# I don't know how to name them in order not to duplicate them!

def _convert_name(name):
    # on Linux / Mac OS X 'foo.PY' is not importable, but on
    # Windows it is. Simpler to do a case insensitive match
    # a better check would be to check that the name is a
    # valid Python module name.
    if os.path.isfile(name) and name.lower().endswith('.py'):
        if os.path.isabs(name):
            rel_path = os.path.relpath(name, os.getcwd())
            if os.path.isabs(rel_path) or rel_path.startswith(os.pardir):
                return name
            name = rel_path
        # on Windows both '\' and '/' are used as path
        # separators. Better to replace both than rely on os.path.sep
        return name[:-3].replace('\\', '.').replace('/', '.')
    return name

def _convert_names(names):
    return [_convert_name(name) for name in names]

main = TestProgram

# end.
